<?php

namespace Drupal\xnttmulti\Plugin\ExternalEntities\StorageClient;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\Database\DatabaseExceptionWrapper;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\external_entities\ExternalEntityInterface;
use Drupal\external_entities\ExternalEntityTypeInterface;
use Drupal\external_entities\Plugin\PluginFormTrait;
use Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface;
use Drupal\external_entities\StorageClient\ExternalEntityStorageClientBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\dbxschema\Database\CrossSchemaConnectionInterface;
use Drupal\xnttmulti\Exception\MultiExternalEntityException;
use Drupal\xnttmulti\Entity\ReadOnlyExternalEntitySubType;
use Symfony\Component\Yaml\Yaml;

/**
 * External entities multiple storage client.
 *
 * @ExternalEntityStorageClient(
 *   id = "xnttmulti",
 *   label = @Translation("Multiple storages"),
 *   description = @Translation("Generates a combined record from multiple storage sources.")
 * )
 */
class Multi extends ExternalEntityStorageClientBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * Aggregation modes.
   *
   * - AGGREGATE_ALL: aggregate all sources together regardless their group
   *   prefix.
   * - AGGREGATE_BY_GROUP_AND_COMMON: aggregate only source with the same group
   *   prefix and sources without prefix. If an entity identifier does not match
   *   and group prefix, the identifier is considered invalid.
   * - AGGREGATE_BY_GROUP_PLUS_DEFAULT: aggregate only source with the same
   *   group prefix and treate sources without prefix as a default group when
   *   the entity identifier does not match any group prefix.
   */
  public const AGGREGATE_ALL                   = 'all';
  public const AGGREGATE_BY_GROUP_AND_COMMON   = 'bygcom';
  public const AGGREGATE_BY_GROUP_PLUS_DEFAULT = 'bygdef';


  /**
   * Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The external storage client manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $storageClientManager;

  /**
   * Constructs a multiple external storage object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface $response_decoder_factory
   *   The response decoder factory service.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $storage_client_manager
   *   The external storage client manager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   Logger service (xnttmulti).
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    ResponseDecoderFactoryInterface $response_decoder_factory,
    PluginManagerInterface $storage_client_manager,
    MessengerInterface $messenger,
    LoggerChannelInterface $logger
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $string_translation,
      $response_decoder_factory
    );
    // Services injection.
    $this->storageClientManager = $storage_client_manager;
    $this->messenger = $messenger;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('external_entities.response_decoder_factory'),
      $container->get('plugin.manager.external_entities.storage_client'),
      $container->get('messenger'),
      $container->get('logger.channel.xnttmulti')
    );
  }

  /**
   * Provides the list of available plugins.
   *
   * @param bool $reset
   *   Set to TRUE to reload the list of plugin (clears the static cache).
   *   Default: FALSE (use the static cache).
   *
   * @return array
   *   An array with the following keys containing arrays keyed by plugin ids:
   *   - plugins: an array of storage client plugin objects;
   *   - definitions: an array of plugin descriptions;
   *   - options: an array ready to be used for "radios" form elements.
   */
  protected function getOtherAvailablePlugins(bool $reset = FALSE) {
    static $storage_client_options = [];
    $hash_key = spl_object_hash($this);
    if (empty($storage_client_options[$hash_key]) || $reset) {
      $storage_client_options[$hash_key] = ['options' => [], 'plugins' => [], ];
      // Get all storage definitions to provide a list of storage plugins.
      $all_storage_clients = $this->storageClientManager->getDefinitions();
      foreach ($all_storage_clients as $storage_client_id => $definition) {
        // Exclude this plugin from the list.
        if ($this->getPluginId() == $storage_client_id) {
          continue;
        }
        $config = [
          '_external_entity_type' => $this->externalEntityType,
        ];

        $storage_client = $this->storageClientManager->createInstance(
          $storage_client_id,
          $config
        );
        $storage_client_options[$hash_key]['plugins'][$storage_client_id] =
          $storage_client
        ;
        $storage_client_options[$hash_key]['definitions'][$storage_client_id] =
          $definition
        ;
        $storage_client_options[$hash_key]['options'][$storage_client_id] =
          $storage_client->getLabel()
        ;
      }
      asort($storage_client_options[$hash_key]['options'], SORT_NATURAL | SORT_FLAG_CASE);
      $storage_client_options[$hash_key]['options'] =
        ['' => 'Remove']
        + $storage_client_options[$hash_key]['options']
      ;
    }
    return $storage_client_options[$hash_key];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'storages' => [],
    ];
  }

  /**
   * Initialize externalEntityType member if not set.
   */
  protected function initExternalEntityType(FormStateInterface $form_state) {
    // Get current external entity type if not set.
    if (empty($this->externalEntityType)) {
      $state_data = $form_state->getStorage();
      $xntt_type_id = $state_data['machine_name.initial_values']['id'];
      if (!empty($xntt_type_id)) {
        $xntt_type = \Drupal::entityTypeManager()
          ->getStorage('external_entity_type')
          ->load($xntt_type_id)
        ;
        $this->externalEntityType = $xntt_type;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $this->initExternalEntityType($form_state);

    // Get the list of available storage plugins.
    $storage_clients = $this->getOtherAvailablePlugins();

    $form['storages'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Multiple storage settings'),
    ];
    $form['storages']['description'] = [
      '#type' => '#markup',
      '#markup' =>
        '<div>'
        . $this->t(
          'You can specify one or more storage client. Their order is important: the first one is loaded first.'
        )
        . '</div>'
      ,
    ];
    $form['storages']['aggr'] = [
      '#type' => 'radios',
      '#title' => $this->t('Aggregation mode'),
      '#options' => [
        static::AGGREGATE_ALL =>
          $this->t('Aggregate all'),
        static::AGGREGATE_BY_GROUP_AND_COMMON =>
          $this->t('Aggregate by group prefixes with unprefixed storage clients in common'),
        static::AGGREGATE_BY_GROUP_PLUS_DEFAULT =>
          $this->t('Aggregate by group prefixes and treate unprefixed storage clients as a default group'),
      ],
      '#default_value' =>
        $form_state->get('aggr')
        ?? $this->configuration['aggr']
        ?? static::AGGREGATE_ALL,
      '#required' => TRUE,
    ];

    // Get storages from current form or from config.
    $storages =
      $form_state->get('storages')
      ?? $this->configuration['storages']
      ?? []
    ;
    $storage_count = $form_state->get('storage_count') ?? count($storages);
    // We must have at least one storage.
    if (empty($storage_count)) {
      $storage_count = 1;
    }

    $storage_index = 0;
    // Make sure we work on a non-empy list of plugins.
    if (!empty($storage_clients['options'])) {
      for ($i = 0; $i < $storage_count; ++$i) {
        $default_client = $storages[$i]['storage_client'] ?? '';
        // @todo: fix issue when changing plugin: config corresponds to previous
        // plugin instead of the selected one. Maybe check
        // validateConfigurationForm().
        $client_config =
          $storages[$i]['config']
          ?? []
        ;

        // @todo: fix form issue "invalid plugin selection".
        // Issue: Drupal form API will auto-check the wrong radiobox on AJAX
        // reload. If we have 3 storage clients and remove the middle one, the
        // last one will become the "second" item but "Remove" radio will be
        // checked while the displayed settings will correspond to the real
        // last plugin and the 3rd one will remain as an empty plugin but
        // without "Remove" being selected.

        // Only display non-empty clients unless it is the last one.
        if ((!empty($default_client)) || ($i == $storage_count - 1)) {
          if (empty($client_config)
              && !empty($default_client)
              && !empty($this->externalEntityType)
              && ($this->externalEntityType instanceof ExternalEntityTypeInterface)
              && ($this->externalEntityType->getStorageClientId() == $default_client)
          ) {
            $client_config = $this->externalEntityType->getStorageClientConfig();
          }

          $form['storages'][$storage_index] = [
            '#type' => 'details',
            '#title' => $this->t(
              'Storage client #@index',
              ['@index' => $storage_index+1]
            ),
            '#open' => TRUE,
          ];
          $form['storages'][$storage_index]['storage_client'] = [
            '#type' => 'radios',
            '#title' => $this->t('Select a storage client.'),
            '#options' => $storage_clients['options'],
            '#default_value' => $default_client,
            // Only require first element.
            '#required' => (0 === $storage_index),
            '#ajax' => [
              'callback' => [
                get_class($form_state->getFormObject()),
                'buildAjaxStorageClientConfigForm',
              ],
              'wrapper' => 'external-entities-storage-client-config-form',
              'method' => 'replace',
              'effect' => 'fade',
            ],
            '#attributes' => ['autocomplete' => 'off',],
          ];

          // Get client config.
          if (!empty($default_client)) {
            $form['storages'][$storage_index]['groups'] = [
              '#type' => 'textfield',
              '#title' => $this->t('Group prefix(es)'),
              '#description' => $this->t('Optional: you can set a group prefix for group aggregation (see "Aggregation mode" setting above) or to strip a prefix from the identifier provided to the storage client. Multiple prefixes can be specified separated by semicolons (;).'),
              '#default_value' => implode(';', $storages[$i]['groups'] ?? []),
              '#attributes' => [
                'id' => 'edit-storages-' . $storage_index . '-groups',
              ],
            ];
            $form['storages'][$storage_index]['group_prefix_strip'] = [
              '#type' => 'checkbox',
              '#title' => $this->t('Strip group prefix from external entity identifier for storage client identifiers'),
              '#description' => $this->t('When using a group name, external entity identifiers is prefixed by the group name while the storage client can get the unprefixed identifier if the checkbox is checked.'),
              '#default_value' => $storages[$i]['group_prefix_strip'] ?? FALSE,
              '#states' => [
                'invisible' => [
                  ':input[id="edit-storages-' . $storage_index . '-groups"]' => ['value' => ''],
                ],
              ],
            ];

            // Attach the storage client plugin configuration form.
            $form['storages'][$storage_index]['config'] = [];
            $storage_client = $storage_clients['plugins'][$default_client];
            $storage_client->setConfiguration($client_config);

            $storage_client_form_state = SubformState::createForSubform(
              $form['storages'][$storage_index]['config'],
              $form,
              $form_state
            );
            $form['storages'][$storage_index]['config'] =
              $storage_client->buildConfigurationForm(
                $form['storages'][$storage_index]['config'],
                $storage_client_form_state
              )
            ;
            $form['storages'][$storage_index]['config'] += [
              '#type' => 'fieldset',
              '#title' => $this->t(
                ':plugin Settings',
                [':plugin' => $storage_clients['options'][$default_client],]
              ),
              '#open' => TRUE,
            ];
            $description =
              $storage_clients['definitions'][$default_client]['description']
            ;
            if (!empty($description)) {
              $form['storages'][$storage_index]['config'] +=
                ['#description' => $description]
              ;
            }

            // The use of override or other identifier field can not work with
            // the first storage client as no field has been loaded for it.
            if (0 !== $storage_index) {
              $form['storages'][$storage_index]['override'] = [
                '#type' => 'checkbox',
                '#title' => $this->t('Override previous field values'),
                '#description' => $this->t('If set, entity field values provided by this plugin will override existing values with the same field name provided by previous plugins.'),
                '#default_value' => $storages[$i]['override'] ?? TRUE,
              ];

              $form['storages'][$storage_index]['id_field'] = [
                '#type' => 'textfield',
                '#title' => $this->t('Field name to use as identifier for this storage client.'),
                '#description' => $this->t('Leave empty to use the default id provided. Otherwise, any text/numeric field provided by previous storage plugins can be used here.'),
                '#default_value' => $storages[$i]['id_field'] ?? NULL,
              ];
            }

            $form['storages'][$storage_index]['readonly'] = [
              '#type' => 'checkbox',
              '#title' => $this->t('Read only (no create/update) from client.'),
              '#default_value' => $storages[$i]['readonly'] ?? FALSE,
            ];

          }
          else {
            $form['storages'][$storage_index]['config'] = [
              '#type' => 'container',
            ];
          }
          ++$storage_index;
        }
      }
    }
    $storage_count = $storage_index;
    $form_state->set('storage_count', $storage_count);
    $form['storages']['add_storage'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add a storage'),
      '#name' => 'add_storage',
      '#ajax' => [
        'callback' => [
          get_class($form_state->getFormObject()),
          'buildAjaxStorageClientConfigForm'
        ],
        'wrapper' => 'external-entities-storage-client-config-form',
        'method' => 'replace',
        'effect' => 'fade',
      ],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * Note: there is no ::submitConfigurationForm because External Entities
   * plugin system does not use it. All is done here.
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    $this->initExternalEntityType($form_state);

    $update_count = FALSE;
    // Check for Ajax events.
    if ($trigger = $form_state->getTriggeringElement()) {
      switch ($trigger['#name']) {
        case 'add_storage':
          $storage_count = $form_state->get('storage_count') + 1;
          $form_state->set('storage_count', $storage_count);
          $form_state->setRebuild(TRUE);
          $update_count = TRUE;
          break;

        default:
      }
    }

    // Storages.
    $storages = $form_state->getValue('storages');
    $storage_clients = $this->getOtherAvailablePlugins();
    $storage_client_configs = [];
    foreach ($storages as $storage_index => $storage) {
      // Skip non-client items or empty clients.
      if (!is_int($storage_index) || empty($storage['storage_client'])) {
        continue;
      }
      $storage_client_id = $storage['storage_client'];
      // Get plugin.
      $storage_client = $storage_clients['plugins'][$storage_client_id] ?? NULL;
      if (empty($storage_client)) {
        $update_count = TRUE;
        continue;
      }
      $storage_client_form_state = SubformState::createForSubform(
        $form['storages'][$storage_index]['config'],
        $form,
        $form_state
      );
      $storage_client->validateConfigurationForm(
        $form['storages'][$storage_index]['config'],
        $storage_client_form_state
      );

      $storage_client_configs[] = [
        'storage_client' => $storage_client_id,
        'id_field' => $storage['id_field'] ?? '',
        'override' => $storage['override'] ?? FALSE,
        'readonly' => $storage['readonly'] ?? TRUE,
        'groups'   => array_filter(
          preg_split('/\s*;\s*/', trim($storage['groups'] ?? '')),
          'strlen'
        ),
        'group_prefix_strip' => $storage['group_prefix_strip'] ?? FALSE,
        'config' => $storage_client->getConfiguration(),
      ];
    }

    // Update storage count according to "Add storage" button or used "Remove"
    // radios.
    if ($update_count) {
      $form_state->set(
        'storage_count',
        $storage_count ?? count($storage_client_configs)
      );
      $form_state->setRebuild(TRUE);
    }

    //@todo: maybe try to set checked radios here for plugin selection to fix
    // issue "invalid plugin selection".
    // Tryed in $form without success.
    //       $form['storages'][$form_index]['storage_client']['#value'] = $storage_client_id;
    //       $form['storages'][$form_index]['storage_client']['#default_value'] = $storage_client_id;
    //       $form['storages'][$form_index]['storage_client'][$storage_client_id]['#default_value'] = $storage_client_id;
    //       $form['storages'][$form_index]['storage_client'][$storage_client_id]['#checked'] = TRUE;

    $form_state->setValue('storages', $storage_client_configs);
    // Save state for form update when adding items.
    $form_state->set('storages', $storage_client_configs);

    // Aggregation settings.
    $form_state->setValue('aggr', $storages['aggr'] ?? static::AGGREGATE_ALL);

    $this->setConfiguration($form_state->getValues());
  }

  /**
   * Returns the ordered list of configured plugin for current instance.
   */
  public function getPlugins($reload = FALSE) {
    // Static cache.
    static $plugins = [];
    $hash_key = spl_object_hash($this);
    if (!isset($plugins[$hash_key]) || $reload) {
      $plugins[$hash_key] = [];
      $storages = $this->getConfiguration()['storages'] ?? [];
      foreach ($storages as $storage_settings) {
        // Sets externalEntityType member if missing in plugin config.
        if (!empty($this->externalEntityType)
            && ($this->externalEntityType instanceof ExternalEntityTypeInterface)
        ) {
          // Create a "read-only" clone of the real external entity type.
          $client_xntt_type = ReadOnlyExternalEntitySubType::createFromParent(
            $this->externalEntityType
          );
          // Redefines id mapping if needed.
          if (!empty($storage_settings['id_field'])) {
            $field_mapper = $client_xntt_type->getFieldMapper();
            $field_mappings = $field_mapper->getFieldMappings();
            $field_mappings['id'] = ['value' => $storage_settings['id_field']];
            // @todo: check if this works properly.
            $client_xntt_type->set('field_mappings', $field_mappings);
          }
          $storage_settings['config'] +=
            ['_external_entity_type' => $client_xntt_type,]
          ;
        }

        $plugins[$hash_key][] = $this->storageClientManager->createInstance(
          $storage_settings['storage_client'],
          $storage_settings['config']
        );
      }
    }
    return $plugins[$hash_key];
  }

  /**
   * Filter aggregation plugin.
   *
   * Only return storage plugins that are relevant to the provided entity
   * identifier. It takes into account if default group ('') is considered as a
   * separate group or not. Since some plugins may use alternative fields as
   * identifiers, they will not be filtered here but they should be fitered
   * afterward, from the related identifier field values.
   *
   * @param $id
   *   Identifier value of current external entity considered.
   */
  protected function filterAggregationPluginsForId($id) {
    $plugins = $this->getPlugins();
    $config = $this->getConfiguration();
    $aggregate = $config['aggr'] ?? static::AGGREGATE_ALL;
    $storage_config = $config['storages'];

    // If aggregating by groups plus default, filter plugins to use.
    if (static::AGGREGATE_BY_GROUP_PLUS_DEFAULT == $aggregate) {
      $group_plugins = [];
      $default_plugins = [];
      // List default storage plugins and check if there is a group match with
      // a group to see if we use default or a group.
      foreach ($plugins as $plugin_index => $plugin) {
        $groups = $storage_config[$plugin_index]['groups'] ?? [];
        if (empty($groups)) {
          $default_plugins[$plugin_index] = $plugin;
        }
        else {
          $re = '/^\Q' .implode('\E|^\Q', $groups) . '\E/';
          if (1 === preg_match($re, $id)) {
            $group_plugins[$plugin_index] = $plugin;
          }
        }
      }
      $aggr_plugins = $group_plugins ?: $default_plugins;
    }
    else {
      // Otherwise, work on all plugins.
      $aggr_plugins = $plugins;
    }

    return $aggr_plugins;
  }

  /**
   * Returns the field name used as identifier.
   */
  public function getIdField($plugin_config = NULL) {
    static $id_fields = [];
    $hash_key =
      spl_object_hash($this)
      . ($plugin_config['id_field']['value'] ?? '')
    ;
    if (!isset($id_field[$hash_key])) {
      if (!empty($plugin_config['id_field']['value'])) {
        $id_field = $plugin_config['id_field']['value'];
      }
      else {
        $field_mapper = $this->externalEntityType->getFieldMapper();
        $id_field = $field_mapper->getFieldMapping('id')['value'];
      }
      $id_fields[$hash_key] = $id_field;
    }
    return $id_fields[$hash_key];
  }

  /**
   * {@inheritdoc}
   */
  public function delete(ExternalEntityInterface $entity) {
    $storage_config = $this->getConfiguration()['storages'];
    $field_mapper = $this->externalEntityType->getFieldMapper();
    $id_field = $field_mapper->getFieldMapping('id')['value'];
    $aggregate = $this->getConfiguration()['aggr'] ?? static::AGGREGATE_ALL;
    // Filter aggregation plugins.
    $aggr_plugins = $this->filterAggregationPluginsForId($entity->id());

    foreach ($aggr_plugins as $plugin_index => $plugin) {
      if (empty($storage_config[$plugin_index]['readonly'])) {
        // Get sub-id.
        $sub_id = $entity->id();
        if (!empty($storage_config[$plugin_index]['id_field'])) {
          $sub_id = $entity->{$storage_config[$plugin_index]['id_field']};
        }
        if (empty($sub_id)) {
          // No sub-id available, skip current storage plugin.
          continue 1;
        }

        // Check for groups.
        $groups = $storage_config[$plugin_index]['groups'] ?? [];

        // Skip aggregation if no group match the ID.
        if ((static::AGGREGATE_BY_GROUP_AND_COMMON == $aggregate)
            && !empty($groups)
        ) {
          $group_match = FALSE;
          foreach ($groups as $group) {
            // Note: we can use sub-id for checking group.
            if (strncmp($sub_id, $group, strlen($group)) === 0) {
              $group_match = TRUE;
              break;
            }
          }
          if (!$group_match) {
            // No match, process next storage plugin.
            continue 1;
          }
        }
        // Strip group prefix if needed.
        if (!empty($storage_config[$plugin_index]['group_prefix_strip'])) {
          $id_field = $this->getIdField($storage_config[$plugin_index]);
          // Get current identifier value.
          $original_id = $sub_id = $entity->get($id_field)->value;
          // Remove group prefixes.
          $re = '/^\Q' .implode('\E|^\Q', $groups) . '\E/';
          $sub_id = preg_replace($re, '', $sub_id);
          // Update identifier value.
          $entity->set($id_field, $sub_id, FALSE);
          $plugin->delete($entity);
          // Put back previous value.
          $entity->set($id_field, $original_id, FALSE);
        }
        else {
          $plugin->delete($entity);
        }
      }
    }

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(array $ids = NULL) {
    $data = [];
    // Load each entity.
    // @todo: load multiple entities by batch and then merge each parts instead
    //   of loading them 1 by 1.
    if (!empty($ids) && is_array($ids)) {
      $aggregate = $this->getConfiguration()['aggr'] ?? static::AGGREGATE_ALL;
      $storage_config = $this->getConfiguration()['storages'];
      $field_mapper = $this->externalEntityType->getFieldMapper();
      foreach ($ids as $id) {
        // Filter aggregation plugins.
        $aggr_plugins = $this->filterAggregationPluginsForId($id);

        $entity = [];
        foreach ($aggr_plugins as $plugin_index => $plugin) {
          // Load and merge values from each storage client.
          $id_field = $this->getIdField($storage_config[$plugin_index]);
          $sub_id = $entity[$id_field] ?? $id;
          if (empty($sub_id)) {
            // No sub-id available, skip current storage plugin.
            continue 1;
          }

          // Check for groups.
          $groups = $storage_config[$plugin_index]['groups'] ?? [];

          // Skip aggregation if no group match the ID.
          if ((static::AGGREGATE_BY_GROUP_AND_COMMON == $aggregate)
              && !empty($groups)
          ) {
            $group_match = FALSE;
            foreach ($groups as $group) {
              // Note: we can use sub-id for checking group.
              if (strncmp($sub_id, $group, strlen($group)) === 0) {
                $group_match = TRUE;
                break;
              }
            }
            if (!$group_match) {
              // No match, process next storage plugin.
              continue 1;
            }
          }

          // If groups are used, check for id strippping.
          if (!empty($storage_config[$plugin_index]['group_prefix_strip'])) {
            // Strip group prefixes.
            $re = '/^\Q' .implode('\E|^\Q', $groups) . '\E/';
            $sub_id = preg_replace($re, '', $sub_id);
          }

          if (empty($sub_id)) {
            // Stripping led to an empty sub-id, skip current storage plugin.
            continue 1;
          }

          $plugin_entity_data = $plugin->load($sub_id) ?? [];

          // Put back (first) prefix if it was stripped.
          if (!empty($storage_config[$plugin_index]['group_prefix_strip'])) {
            // Strip group prefixes.
            $plugin_entity_data[$id_field] = $groups[0] . $sub_id;
          }

          if (!empty($storage_config[$plugin_index]['override'])
              && $storage_config[$plugin_index]['override']
          ) {
            // Override any field if needed.
            $entity = NestedArray::mergeDeep($entity, $plugin_entity_data);
          }
          else {
            // No override of fields already set.
            // @todo: maybe add an option to only override empty strings?
            //   It might be done by removing empty string values from $entity.
            $entity = NestedArray::mergeDeep($plugin_entity_data, $entity);
          }
        }

        // Keep non-empty entities.
        if (!empty($entity)) {
          $data[$id] = $entity;
        }
      }
    }
    return $data;
  }

  /**
   * Loads one entity.
   *
   * @param mixed $id
   *   The ID of the entity to load.
   *
   * @return array|null
   *   A raw data array, NULL if no data returned.
   */
  public function load($id) {
    $data = $this->loadMultiple([$id]);
    return array_key_exists($id, $data)
      ? $data[$id]
      : NULL
    ;
  }

  /**
   * {@inheritdoc}
   */
  public function save(ExternalEntityInterface $entity) {
    $storage_config = $this->getConfiguration()['storages'];
    $field_mapper = $this->externalEntityType->getFieldMapper();
    $id_field = $field_mapper->getFieldMapping('id')['value'];
    $aggregate = $this->getConfiguration()['aggr'] ?? static::AGGREGATE_ALL;
    // Filter aggregation plugins.
    $aggr_plugins = $this->filterAggregationPluginsForId($entity->id());
    foreach ($aggr_plugins as $plugin_index => $plugin) {
      if (empty($storage_config[$plugin_index]['readonly'])) {
        // Get sub-id.
        $sub_id = $entity->id();
        if (!empty($storage_config[$plugin_index]['id_field'])) {
          $sub_id = $entity->{$storage_config[$plugin_index]['id_field']};
        }
        if (empty($sub_id)) {
          // No sub-id available, skip current storage plugin.
          continue 1;
        }

        // Check for groups.
        $groups = $storage_config[$plugin_index]['groups'] ?? [];

        // Skip aggregation if no group match the ID.
        if ((static::AGGREGATE_BY_GROUP_AND_COMMON == $aggregate)
            && !empty($groups)
        ) {
          $group_match = FALSE;
          foreach ($groups as $group) {
            // Note: we can use sub-id for checking group.
            if (strncmp($sub_id, $group, strlen($group)) === 0) {
              $group_match = TRUE;
              break;
            }
          }
          if (!$group_match) {
            // No match, process next storage plugin.
            continue 1;
          }
        }
        // Strip group prefix if needed.
        if (!empty($storage_config[$plugin_index]['group_prefix_strip'])) {
          $sub_id_field = $id_field;
          // Check for an alternative identifier field.
          if (!empty($storage_config[$plugin_index]['id_field'])) {
            $sub_id_field = $storage_config[$plugin_index]['id_field']['value'];
          }
          // Get current identifier value.
          $original_id = $sub_id = $entity->get($sub_id_field)->value;
          // Remove group prefixes.
          $re = '/^\Q' .implode('\E|^\Q', $groups) . '\E/';
          $sub_id = preg_replace($re, '', $sub_id);
          // Update identifier value.
          $entity->set($sub_id_field, $sub_id, FALSE);
          $plugin->save($entity);
          // Put back previous value.
          $entity->set($sub_id_field, $original_id, FALSE);
        }
        else {
          $plugin->save($entity);
        }
      }
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function query(
    array $parameters = [],
    array $sorts = [],
    $start = NULL,
    $length = NULL,
    bool $reload = FALSE
  ) {
    $entities = [];
    $plugins = $this->getPlugins();
    $storage_config = $this->getConfiguration()['storages'];
    $field_mapper = $this->externalEntityType->getFieldMapper();
    $id_field = $field_mapper->getFieldMapping('id')['value'];
    // @todo: manage pager: each plugin can return $length objects and we may
    // have more than the number of expected items per page.
    // List entities from plugins.
    foreach ($plugins as $plugin_index => $plugin) {
      // Get xntt from plugin.
      $plugin_entities = $plugin->query($parameters, $sorts, $start, $length, $reload);
      // Get plugin groups if some.
      $groups = $storage_config[$plugin_index]['groups'] ?? [];
      if (!empty($groups)) {
        // There are groups, get identifier field where the group should appear.
        $sub_id_field = $id_field;
        // Check for an alternative identifier field.
        if (!empty($storage_config[$plugin_index]['id_field'])) {
          $sub_id_field = $storage_config[$plugin_index]['id_field']['value'];
        }
        // Check if the group prefix is stripped.
        if (!empty($storage_config[$plugin_index]['group_prefix_strip'])) {
          // Add (first) group prefix to ID as it is stripped.
          $unprefixed_plugin_entities = $plugin_entities;
          $plugin_entities = [];
          foreach ($unprefixed_plugin_entities as $key => $value) {
            // Prefix ID field.
            $value[$sub_id_field] = $groups[0] . $value[$sub_id_field];
            // Prefix entity key.
            $plugin_entities[$groups[0] . $key] = $value;
          }
        }
        else {
          // Group prefix is part of the xntt ID.
          // Filter returned id to make sure they match their groups.
          $unfiltered_plugin_entities = $plugin_entities;
          $plugin_entities = [];
          foreach ($unfiltered_plugin_entities as $key => $value) {
            $group_match = FALSE;
            foreach ($groups as $group) {
              if (strncmp($key, $group, strlen($group)) === 0) {
                // Got a match.
                $group_match = TRUE;
                break;
              }
            }
            if (!$group_match) {
              // No match, process next entity.
              continue 1;
            }
            // Matched entity.
            $plugin_entities[$key] = $value;
          }
        }
      }
      // Just add filtered/missing entities.
      $entities += $plugin_entities;
    }
    return $entities;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuery(array $parameters = []) {
    $count = 0;
    $plugins = $this->getPlugins();
    $storage_config = $this->getConfiguration()['storages'];
    // Check aggregation type.
    // Depending on aggregation type, the count may change:
    // - if all is aggregated, then we could rely on the first main plugin that
    //   is supposed to provide all the identifiers and the next plugins are
    //   just expected to complete its entities
    // - if it is aggregated by group, then we need to sum by groups.
    $aggregate = $this->getConfiguration()['aggr'] ?? static::AGGREGATE_ALL;
    if (static::AGGREGATE_ALL != $aggregate) {
      // Xntt are aggregated by group.
      // We can't have a real count (without using IDs to find common entities)
      // so we will just sum everything. The returned count may be
      // over-estimated but not under-estimated.
      foreach ($plugins as $plugin_index => $plugin) {
        $count += $plugin->countQuery($parameters);
      }
    }
    else {
      // List entities with first plugin only.
      $count = $plugins[0]->countQuery($parameters);
    }
    // @todo: we don't handle if more entities are available from next
    // plugins. Should we rely on "count($this->query(...))" and/or add
    // a setting option to let the admin choose how external entities should
    // be counted (first plugin, plugin 'n', sum of all plugins, count on
    // query(),...).
    return $count;
  }

}
